public variant ULongTree ['a]
  | Empty
  | Leaf
      k : ulong
      elt : 'a
  | Branch
      p : uint
      m : uint
      t1 : ULongTree ['a]
      t2 : ULongTree ['a]

  | BranchDeep
      p : ulong
      m : ulong
      t1 : ULongTree ['a]
      t2 : ULongTree ['a]

  public IsEmpty : bool
    get
      this is Empty


  public Get (k : ulong) : 'a
    def loop (_)
      | Leaf (k', r) when k' == k => r
      | Branch (_, m, l, r) =>
        if (k :> uint & m == 0) loop (l) else loop (r)
      | BranchDeep (_, m, l, r) =>
        if (k & m == 0) loop (l) else loop (r)

      | _ =>
        throw ArgumentException ("Get: key not found")
    loop (this)

  public TryGetValue (k : ulong, v : out 'a) : bool
    def loop (_)
      | Leaf (k', v') when k' == k => v = v'; true
      | Branch (_, m, l, r) =>
        if ((k :> uint & m) == 0) loop (l) else loop (r)
      | BranchDeep (_, m, l, r) =>
        if ((k & m) == 0) loop (l) else loop (r)
      | _ => false

    loop(this)


  public Add (k : ulong, v : 'a) : ULongTree ['a]
   unchecked
    def join (p0 : ulong, t0, p1, t1)
      def p = p0 ^ p1
      def m = (p & ((-(p :> long)) :> ulong))
      if( (m & 0xFFFFFFFF) == 0 )
        if ((p0 & m) == 0)
          BranchDeep (p0 & (m - 1), m, t0, t1)
        else
          BranchDeep (p0 & (m - 1), m, t1, t0)
      else
        if ((p0 & m) == 0)
          Branch ( (p0 & (m - 1)) :> int, m :>int, t0, t1)
        else
          Branch ( (p0 & (m - 1)) :> int, m :>int, t1, t0)

    def ins (t)
      | Leaf (k', _) when k' == k => Leaf (k, v)

      | Leaf (k', _) => join (k, Leaf (k, v), k', t)

      | Branch (p, m, t0, t1) as t =>
        if (k & (m - 1) == p)
          if (k & m == 0)
            Branch (p, m, ins (t0), t1)
          else
            Branch (p, m, t0, ins (t1))
        else
          join (k, Leaf (k, v), p, t)
      | BranchDeep (p, m, t0, t1) as t =>
        if (k & (m - 1) == p)
          if (k & m == 0)
            BranchDeep (p, m, ins (t0), t1)
          else
            BranchDeep (p, m, t0, ins (t1))
        else
          join (k, Leaf (k, v), p, t)

    match(this)
      | Leaf (k', _) when k' == k \
      | Empty => Leaf (k, v)

      | Leaf (k', _) => join (k, Leaf (k, v), k', t)

      | Branch (p, m, t0, t1) as t =>
        if (k & (m - 1) == p)
          if (k & m == 0)
            Branch (p, m, ins (t0), t1)
          else
            Branch (p, m, t0, ins (t1))
        else
          join (k, Leaf (k, v), p, t)




